#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define EXPECT(a) if (gettoken(in,token)==0) return; \
                  if (strcmp(token,a)!=0) { \
                  printf("Parse Error: Expecting '%s' but got '%s'.\n",a,token); \
                  return; }

void pushback(char *token);
int gettoken(FILE *in, char *token);
void parse_block(FILE *in, FILE *out, int indent);
void parse_expr(FILE *in, FILE *out, char break_char);

void spacer(FILE *out, int indent)
{
int t;

  for (t=0; t<indent; t++) putc(' ',out);
}

void print_err(char *s, char *token)
{
  printf("Parse Error: Expected '%s' but got '%s'.\n",s,token);
}

int token_needs_space(char *token)
{
const char *s[] = { "-","+","/","*","&","|","&&","||", 0 };
int t;

  t=0;
  while(s[t]!=0)
  {
    if (strcmp(s[t++],token)==0) return 1;
  }

  return 0;
}

void parse_expr_paren(FILE *in, FILE *out)
{
  fprintf(out,"(");
  parse_expr(in,out,')');
  fprintf(out,")");
}

void parse_expr_square(FILE *in, FILE *out)
{
  fprintf(out,"[");
  parse_expr(in,out,']');
  fprintf(out,"]");
}

/*
void parse_expr_funct(FILE *in, FILE *out)
{
  fprintf(out,"(");
  parse_expr(in,out,')');
  fprintf(out,")");
}
*/

void parse_expr(FILE *in, FILE *out, char break_char)
{
char token[1024];
int eq=0;

#ifdef DEBUG
printf("Enter parse_expr()\n");
#endif

  while(gettoken(in,token)!=0)
  {
#ifdef DEBUG
printf("expr='%s'\n",token);
#endif

    if (token[0]==';') break;
    if (token[0]==break_char) break;

#ifdef PROPTARI
    if (strcmp(token,"ram")==0)
    {
      fprintf(out,"%s[remap(",token);
      EXPECT("[")
      // spacer(out,indent);
      parse_expr(in,out,']');
      fprintf(out,")]");
      continue;
    }
#endif

    if (strcmp(token,"(")==0)
    {
      parse_expr_paren(in,out);
    }
      else
    if (strcmp(token,"[")==0)
    {
      parse_expr_square(in,out);
    }
      else
    if (strcmp(token,"=")==0)
    {
      fprintf(out," := ");
      eq++;
      if (eq>1)
      {
        printf("Parse Error: Too many '=' in expression?\n");
        return;
      }
    }
      else
    {
      if (token_needs_space(token))
      { fprintf(out," %s ",token); }
        else
      { fprintf(out,"%s",token); }
    }
  }

#ifdef DEBUG
printf("Leave parse_expr()\n");
#endif
}

void parse_cond(FILE *in, FILE *out, int indent)
{
char token[1024];

#ifdef DEBUG
printf("Entering parse_condition()\n");
#endif

  EXPECT("(")

  parse_expr(in,out,')');

  fprintf(out,"\n");

  if (gettoken(in,token)==0) return;
  indent=indent+2;

  if (strcmp(token,"{")==0)
  {
    parse_block(in,out,indent);
  }
    else
  {
    spacer(out,indent);
    fprintf(out,"%s",token);
    parse_expr(in,out,';');
    fprintf(out,"\n");
  }

#ifdef DEBUG
printf("Leaving parse_condition()\n");
#endif
}

void parse_if(FILE *in, FILE *out, int indent)
{
#ifdef DEBUG
printf("Entering parse_if()\n");
#endif

  spacer(out,indent);
  fprintf(out,"if ");
  parse_cond(in,out,indent);

#ifdef DEBUG
printf("Leaving parse_if()\n");
#endif
}

void parse_else(FILE *in, FILE *out, int indent)
{
char token[1024];

#ifdef DEBUG
printf("Entering parse_else()\n");
#endif

  spacer(out,indent);
  if (gettoken(in,token)==0) return;

  if (strcmp(token,"if")==0)
  {
    fprintf(out,"elseif ");
    parse_cond(in,out,indent);
  }
    else
  {
    fprintf(out,"else\n");

    if (strcmp(token,"{")==0)
    {
      parse_block(in,out,indent+2);
    }
      else
    {
      fprintf(out,"%s",token);
      parse_expr(in,out,';');
      fprintf(out,"\n");
    }
  }

#ifdef DEBUG
printf("Leaving parse_else()\n");
#endif
}

void parse_while(FILE *in, FILE *out, int indent)
{
  spacer(out,indent);
  fprintf(out,"repeat while ");
  parse_cond(in,out,indent);
}

void parse_block(FILE *in, FILE *out, int indent)
{
char token[1024];

#ifdef DEBUG
printf("Entering parse_block()\n");
#endif

  while(gettoken(in,token)!=0)
  {
#ifdef DEBUG
printf("token='%s'\n",token);
#endif

    if (strcmp(token,"}")==0) return;

    if (strcmp(token,"if")==0) parse_if(in,out,indent);
      else
    if (strcmp(token,"else")==0) parse_else(in,out,indent);
      else
    if (strcmp(token,"while")==0) parse_while(in,out,indent);
      else
    {
      pushback(token);
      spacer(out,indent);
      // fprintf(out,"%s",token);
      parse_expr(in,out,';');
      fprintf(out,"\n");
    }
  }
}

void parse_func_params(FILE *in, FILE *out)
{
char token[1024];

  EXPECT("(")

  fprintf(out,"(");

  while(gettoken(in,token)!=0)
  {
    // I don't know why I'm even putting this half ass incomplete thing
    // here since I don't need these anyway.. maybe for future additions
    if (strcmp(token,"int")==0 ||
        strcmp(token,"char")==0)
    {
      gettoken(in,token);
    }

    fprintf(out,"%s",token);

    if (strcmp(token,",")==0)
    { continue; }

    if (strcmp(token,")")==0) break;
  }

}

void parse_function(FILE *in, FILE *out)
{
char token[1024];

  // Get function name
  if (gettoken(in,token)==0)
  {
    printf("Parse Error: Expected function name and got '%s'\n",token);
    return;
  }

  fprintf(out,"PUB %s",token);

  parse_func_params(in,out);
  fprintf(out,"\n");

  if (gettoken(in,token)==0) return;

  if (strcmp(token,"{")==0)
  {
    parse_block(in,out,2);
  }
    else
  {
    print_err("{",token);
  }

}

void parse(FILE *in, FILE *out)
{
char token[1024];

  while(1)
  {
    if (gettoken(in,token)==0) break;

    if (strcmp(token,"void")==0)
    {
      parse_function(in,out);
    }
      else
    {
      printf("This doesn't support anything but void functions.\n");
      break;
    }
  }
}

int main(int argc, char *argv[])
{
FILE *in,*out;

  printf("c2spin - Copyright 2008 by Michael Kohn (mike@mikekohn.net)\n");
  printf("http://www.mikekohn.net/\n");
  printf("Version: 2008-Feb-03\n\n");

  if (argc!=3)
  {
    printf("Usage: c2spin <infile.c> <outfile.spin>\n\n");
    exit(0);
  }

  in=fopen(argv[1],"rb");
  if (in==0)
  {
    printf("Could not open input file %s\n",argv[1]);
    exit(1);
  }

  out=fopen(argv[2],"wb");
  if (out==0)
  {
    printf("Could not open output file %s\n",argv[2]);
    fclose(in);
    exit(1);
  }

  fprintf(out,"\n");

  parse(in,out);

  fprintf(out,"\n\n");

  return 0;
}

